#!/usr/bin/python

import rclpy
from rclpy.node import Node
from std_msgs.msg import Float64
from geographic_msgs.msg import GeoPoint  # GeoPoint.latitude, GeoPoint.longitude, GeoPoint.altitude
from time import time
from datetime import datetime
import getpass
import os
import csv


class PointLogger(Node):
    def save(self, data_1, data_2):        
        with open(self.FILENAME, 'a', newline='') as f:
            writer = csv.writer(f)
            now = datetime.now() # current date and time
            writer.writerow([now.strftime("%m/%d/%Y, %H:%M:%S"), *data_1, data_2])

    def update(self):
        """
        We are looking for the last pair of messages that fits the specified maximum interval between records.
        In any case, we clean the buffer of read messages
        """
        if len(self.data) == 2:
            time_pos = self.data["position"][-1]
            time_dth = self.data["depth"][-1]

            if abs(time_pos - time_dth) <= self.MAX_DELTA_TIME:
                lat, lon = self.data["position"][:2]
                dth = self.data["depth"][0]
                if not(any(x is None for x in [lat ,lon, dth])):
                    self.save((lat, lon), dth)
                    self.clear_data()
                else:
                    return

    def __init__(self):
        super().__init__('point_logger')

        self.data = {}

        # geting directory for csv file
        self.declare_parameter('directory', rclpy.Parameter.Type.STRING)
        # geting delays
        self.declare_parameter('max_delta_time', rclpy.Parameter.Type.DOUBLE)
        self.declare_parameter('delay_between_writes', rclpy.Parameter.Type.DOUBLE)

        # connect to gps topic
        self.pos =  self.create_subscription(
            GeoPoint, 
            "gga", 
            self.set_position, 1
        )
        # connect to depth topic
        self.create_subscription(
            Float64, 
           "dbt", 
            self.set_depth, 1
        )

        # make absolute path to dir and make dir if need
        path = self.get_declare_parameter("directory", lambda x: x, "directory not specified")
        # если указан не абсолютный путь, то задаём папку в директории пользователя
        if not path.startswith("/home"):
            # с проверкой / в начале не абсолютного пути, если его там нет то добавляем
            path = f"/home/{getpass.getuser()}{'' if path.startswith('/') else '/'}{path}"
        # добавляем / в конец, если его там нет то добавляем
        path = f"{path}{'' if path.endswith('/') else '/'}"
        # создаём папку, если она не существует
        if not os.path.isdir(path):
            os.makedirs(path)
        self.FILENAME = f"{path}{datetime.today().strftime('%Y-%m-%d %H:%M:%S')}.csv"

        self.MAX_DELTA_TIME = self.get_declare_parameter("max_delta_time", lambda x: x > 0, "max_delta_time not more than zero")

        self.get_logger().info(f"point logger ready to write logs to file {self.FILENAME}")

    def set_position(self, msg):
        self.data["position"]=(msg.latitude, msg.longitude, time())
        self.update()
    
    def set_depth(self, msg):
        self.data["depth"]=(msg.data, time())
        self.update()

    def get_declare_parameter(self, name, checker, error_text):
        value = self.get_parameter(name).value
        if not checker(value):
            self.get_logger().error(error_text)
            raise Exception(error_text)
        return value
    
    def clear_data(self):
        self.data = {}  # self.data = [ DataLog, DataLog, ... ]


class DataLog:
    def __init__(self, name, data, time):
        self.__name = name
        self.__data = data
        self.__time = time
    
    def get_name(self):
        return self.__name

    def get_data(self):
        return self.__data

    def get_time(self):
        return self.__time


def main():
    rclpy.init()
    point_logger = PointLogger()
    rclpy.spin(point_logger)
    point_logger.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()