#!/usr/bin/env python3

import rclpy
from rclpy.node import Node
from std_msgs.msg import Bool
from geometry_msgs.msg import Twist
from geographic_msgs.msg import GeoPoint
from nonius_msgs.msg import Track

from geographiclib.geodesic import Geodesic

import math
           

def translate(value, leftMin, leftMax, rightMin, rightMax):
    # Figure out how 'wide' each range is
    leftSpan = leftMax - leftMin
    rightSpan = rightMax - rightMin

    # Convert the left range into a 0-1 range (float)
    valueScaled = float(value - leftMin) / float(leftSpan)

    # Convert the 0-1 range into a value in the right range.
    return rightMin + (valueScaled * rightSpan)

def scale_angle(value):
    return translate(value, -180.0, 180.0, -1.0, 1.0)

class Controller(Node):
    def add_param(self, name, defvalue):
        self.config[name] = self.declare_parameter(name, defvalue).value

    def parameter_callback(self, data):
        for parameter in data:
            if parameter.name.split('.')[1] == self.name:
                self.config[parameter.name.split('.')[2]] = parameter.value
         
    def __init__(self):
        super().__init__('regulator')
        self.declare_parameter('hz', rclpy.Parameter.Type.INTEGER) 
        self.declare_parameter('distance_threshold', rclpy.Parameter.Type.DOUBLE) 
        #self.declare_parameter('normal_speed', rclpy.Parameter.Type.DOUBLE) 
        self.declare_parameter('min_speed_for_cog', rclpy.Parameter.Type.DOUBLE) 
        self.declare_parameter('k_p_angular', rclpy.Parameter.Type.DOUBLE) 
        self.declare_parameter('max_thrust', rclpy.Parameter.Type.DOUBLE) 
        #self.declare_parameter('k_p_linear', rclpy.Parameter.Type.DOUBLE) 
        #self.declare_parameter('linear_regulator_threshold', rclpy.Parameter.Type.DOUBLE) 

        self.publisher = self.create_publisher(Twist, 'twist_command', 10)
        self.update_timer = self.create_timer(1 / self.get_parameter('hz').value, self.update)

        self.create_subscription(GeoPoint, 'gga', self.set_position, 1)
        self.create_subscription(Twist, 'rmca', self.set_state, 1)
        self.create_subscription(Track, 'target_position', self.set_target, 1)
        self.create_subscription(Bool, 'stop_moving', self.stop, 1)

        self.stopped = True
        self.state = None
        self.target = None
        self.position = None
        self.origin = None
 
        self.path = []
        self.current_target = 0
        
        self.get_logger().info('Auto teleoperation started')
        
    def set_position(self, position):
        self.position = position
    
    def set_state(self, state):
        self.state = state
           
    def stop(self, stop):
        self.get_logger().info('Stop regulation')
        self.stopped = True
        self.publisher.publish(Twist())

    def can_move(self):
        return not self.stopped and self.state is not None and len(self.path) > 0 and self.position is not None
    
    def error_track(self, position):
        vec1 = (self.target.lantitude - self.origin.latitude, self.target.longitude - self.origin.longitude)
        vec2 = (position.lantitude - self.origin.latitude, position.longitude - self.origin.longitude)
        return (vec1[0]*vec2[1] - vec1[1]*vec2[0])
    
    def set_target(self, track):
        if (track.mode == "point" or track.mode == "path") and len(track.points) > 0:
            self.path = track.points
            self.current_target = 0
            self.origin = self.position
            self.stopped = False
            self.get_logger().info(f'Set new target of {len(self.path)} points')

    @staticmethod
    def get_distance_and_heading(position, target):
        geod = Geodesic.WGS84
        res = geod.Inverse(position.latitude, position.longitude, target.latitude, target.longitude)

        result = Twist()
        result.linear.x = res['s12']
        result.angular.z = res['azi2']
        return result

    @staticmethod
    def angle_error(target_angle, source_angle):
        result = target_angle - source_angle
        if result > 180:
            result = result - 360  
        return result

    def move_to_point(self, position, target, state, max_thrust, min_speed, distance_threshold, k_p_angular):
        error = Controller.get_distance_and_heading(position, target)
        heading = state.angular.z
        if heading > 180:
            heading = heading - 360  
        self.get_logger().info(f'Distance {error.linear.x}, heading {heading}, bearing {error.angular.z}')
        if error.linear.x < distance_threshold:
            return None

        result = Twist()
        if state.linear.x > min_speed:
            angle_error = Controller.angle_error(error.angular.z, heading)
            value = scale_angle(angle_error)            
            value = k_p_angular * value            
            self.get_logger().info(f'Angle_error {angle_error}, scaled angle {value}')
            result.angular.z = math.copysign(min(abs(max_thrust), abs(value)), value)
            result.linear.x = (max_thrust - abs(result.angular.z))
        else:
            self.get_logger().info('Speed too low, move straight')
            result.linear.x = max_thrust
        return result

    def update(self):
        if not self.can_move():
           return
        result = self.move_to_point(self.position, self.path[self.current_target], self.state,
                 self.get_parameter('max_thrust').value,
                 self.get_parameter('min_speed_for_cog').value,
                 self.get_parameter('distance_threshold').value,
                 self.get_parameter('k_p_angular').value)
             
        if result is None:
            if self.current_target < len(self.path)-1:
                self.current_target = self.current_target + 1
                self.get_logger().info(f'Point {self.current_target}/{len(self.path)} was achieved')   
            else:
                self.path = []
                self.get_logger().info(f'Path finished')   
                self.stopped = True
        else:
            self.publisher.publish(result)


def main():
    rclpy.init()
    controller = Controller()
    rclpy.spin(controller)
    controller.destroy_node()
    rclpy.shutdown()

if __name__ == '__main__':
    main()