#!/usr/bin/env python3

import rclpy
import math
from rclpy.node import Node
from std_msgs.msg import Bool, Int16, String
from geometry_msgs.msg import Twist, Point
from math import acos, degrees
# from nonius_msgs.msg import Track


import math
           

def translate(value, leftMin, leftMax, rightMin, rightMax):
    # Figure out how 'wide' each range is
    leftSpan = leftMax - leftMin
    rightSpan = rightMax - rightMin

    # Convert the left range into a 0-1 range (float)
    valueScaled = float(value - leftMin) / float(leftSpan)

    # Convert the 0-1 range into a value in the right range.
    return rightMin + (valueScaled * rightSpan)

def scale_angle(value):
    return translate(value, -180.0, 180.0, -1.0, 1.0)

class Controller(Node):
    def add_param(self, name, defvalue):
        self.config[name] = self.declare_parameter(name, defvalue).value

    def parameter_callback(self, data):
        for parameter in data:
            if parameter.name.split('.')[1] == self.name:
                self.config[parameter.name.split('.')[2]] = parameter.value
         
    def __init__(self):
        super().__init__('regulator')
        self.declare_parameter('hz', rclpy.Parameter.Type.INTEGER) 
        self.declare_parameter('distance_threshold', rclpy.Parameter.Type.DOUBLE) 
        #self.declare_parameter('normal_speed', rclpy.Parameter.Type.DOUBLE) 
        self.declare_parameter('k_p_angular', rclpy.Parameter.Type.DOUBLE) 
        self.declare_parameter('max_thrust', rclpy.Parameter.Type.DOUBLE)
        self.declare_parameter('move_type', rclpy.Parameter.Type.STRING) 
        #self.declare_parameter('k_p_linear', rclpy.Parameter.Type.DOUBLE) 
        #self.declare_parameter('linear_regulator_threshold', rclpy.Parameter.Type.DOUBLE) 

        self.publisher = self.create_publisher(Twist, 'twist_robot', 10)
        self.update_timer = self.create_timer(1 / self.get_parameter('hz').value, self.update)

        self.create_subscription(Int16, 'heading', self.get_compass, 1)
        self.create_subscription(Point, 'position', self.get_position, 1)
        self.create_subscription(Point, 'goal', self.get_goal, 1)
        self.create_subscription(String, self.get_parameter('move_type').value, self.move_type, 1)

        self.stopped = True
        self.state = None
        self.goal = Point()
        self.position = Point()
        self.origin = None
        self.heading = 0
        self.type_move = 'STOP'
 
        self.path = []
        self.current_target = 0
        
        self.get_logger().info('Auto teleoperation started')
        
    def get_position(self, position):
        self.position = position
    
    def get_goal(self, state):
        self.goal = state
    def move_type(self, msg):
        self.type_move = msg.data
    
    def get_compass(self, angle):
        self.heading = angle.data
           
    def stop(self, stop):
        self.get_logger().info('Stop regulation')
        self.stopped = True
        self.publisher.publish(Twist())

    def can_move(self):
        return True

    def angle_point_error(self, point, goal):
        x1, y1 = point.x, point.y
        x2, y2 = goal.x, goal.y
        dx, dy = x2 - x1, y2 - y1
        ln = (dx**2 + dy**2)**0.5
        # self.get_logger().info(f'x1 {x1} x2 {x2} y1 {y1} y2 {y2}')
        if ln:
            dxn = dx/ln
            dyn = dy/ln
            ang = degrees(acos(dxn))
            if dyn < 0:
                ang = 360 - ang
            return ang
        return 0 
    
    def angle_error(self,target_angle, source_angle):
        result = (source_angle - target_angle) % 360
        if result > 180:
            result -= 360
        # self.get_logger().info(f'source angle {source_angle} target {target_angle} {result}')
        return result

    def move_to_point(self, position, goal, state, max_thrust, distance_threshold, k_p_angular):
        error_dist = ((goal.x - position.x)**2 + (goal.y - position.y)**2)**0.5
        heading = self.heading
        err_angle = self.angle_error(self.angle_point_error(position, goal), heading)
        # self.get_logger().info(f'{err_angle}')

        result = Twist()
        if self.type_move == "MOVE TO": 
            value = k_p_angular * err_angle
            result.angular.z = math.copysign(min(abs(max_thrust), abs(value)), value)
            if abs(err_angle) < 10:
                result.linear.x = max_thrust
            else: result.linear.x = 0.0
            # self.get_logger().info(f'Angle_error {err_angle}, scaled angle {result.angular.z} linear {result.linear.x}')
        # self.get_logger().info(f'{result}')
        elif self.type_move == "MOVE CIRCLE":
            point = [goal.x, goal.y]
            slf = [position.x, position.y]
            course = heading  # -45
            target_dist = goal.z

            # settings
            is_clockwise = True

            zero = [0, 0]

            eval_dist = lambda p1, p2: ((p1[0] - p2[0])**2 + (p1[1] - p2[1])**2)**0.5
            relate = [slf[0] - point[0], slf[1] - point[1]]
            ln = eval_dist(relate, zero)

            if ln > 0:
                norm = [relate[0] / ln, relate[1] / ln]
            else:
                norm = [1, 0]

            ang = degrees(acos(norm[0]))
            if norm[1] < 0:
                ang = 360 - ang
            ang = ang + (+90 if is_clockwise else -90)

            dist = eval_dist(point, slf)
            delta_dist = dist - target_dist

            angle_by_angle = course - ang
            angle_by_dist = delta_dist/target_dist * 90

            result_angle = (angle_by_angle - angle_by_dist) *0.02

            self.get_logger().info(f"{dist=} {relate=} {ln=} {norm=}")
            self.get_logger().info(f"{round(ang)} {round(delta_dist, 2)}")
            self.get_logger().info(f"{round(angle_by_angle)} {round(angle_by_dist)}")
            result.angular.z = math.copysign(min(abs(max_thrust*2), abs(result_angle)), result_angle)
            result.linear.x = 0.075

        elif self.type_move == "STOP":
            pass    
        return result

    def update(self):
        if not self.can_move():
           return
        result = self.move_to_point(self.position, self.goal, self.heading,
                 self.get_parameter('max_thrust').value,
                 self.get_parameter('distance_threshold').value,
                 self.get_parameter('k_p_angular').value)
             
        if result is None:  
            self.stopped = True
        else:
            # if abs(result.angular.z) < 0.6 and abs(result.linear.x) < 0.15:
            self.publisher.publish(result)
            # self.get_logger().info(f'PUB')


def main():
    rclpy.init()
    controller = Controller()
    rclpy.spin(controller)
    controller.destroy_node()
    rclpy.shutdown()

if __name__ == '__main__':
    main()