#! /usr/bin/env python3

import rclpy
from rclpy.node import Node 
from sensor_msgs.msg import CompressedImage
from std_msgs.msg import Float64
import cv2 
from cv_bridge import CvBridge
import numpy as np
from math import radians, cos, sin


class RealsenseToBinary(Node):
    def __init__(self):
        super().__init__('RealOpenCV')

        # Фронтальная камера для опознавания ворот
        self.declare_parameter('topic_camera', rclpy.Parameter.Type.STRING)

        # Коэффициенты для расчёта расстояния
        # см. ~/Desktop/python/calc.py
        self.declare_parameter('k', rclpy.Parameter.Type.DOUBLE)
        self.declare_parameter('k1', rclpy.Parameter.Type.DOUBLE) 
        self.declare_parameter('k2', rclpy.Parameter.Type.DOUBLE)
        # Горизонтальный угол обзора камеры
        self.declare_parameter('fov', rclpy.Parameter.Type.INTEGER)

        # Параметры для выделения маски синий BGR
        self.hsv_min = (90, 0, 0)
        self.hsv_max = (180, 255, 255)
        self.min_countur_size = 300
        self.get_logger().info(f"HSV \n{self.hsv_min} \n {self.hsv_max}")

        # Магические числа
        x1 = 0
        x2 = 510
        y1 = 170
        y2 = 280

        self.roi_corners = np.array([[(x1,y1),  (x2,y1), 
                         (x2,y2), (x1,y2),
                       ]], 
                       dtype=np.int32)

        # Параметры для рассчёта расстояние и угла
        self.k1 = self.get_parameter('k1').value
        self.k2 = self.get_parameter('k2').value
        self.fov = self.get_parameter('fov').value
        # Пареметр апроксимации контура
        self.k = self.get_parameter('k').value

        topic_name = self.get_declare_parameter("topic_camera", lambda x: x, "topic name must be string")
        
        # Топик с отладочным изображением
        self.pub = self.create_publisher(CompressedImage, f"cv/{topic_name}" , 10)
        # Топик с управляющим сигналом
        self.z_pub = self.create_publisher(Float64, f"{topic_name}/z" , 10)
        # self.distance_publish = self.create_publisher(Float64, 'distance', 10)
        
        self.br = CvBridge() 
        self.subscription = self.create_subscription(CompressedImage, topic_name, self.img_callback, 10)
        self.get_logger().info(f"CV Start {topic_name}")
    
    def get_declare_parameter(self, name, checker, error_text):
        value = self.get_parameter(name).value
        if not checker(value):
            self.get_logger().error(error_text)
            raise Exception(error_text)
        return value

    def float_to_float64(self, msg):
        v = Float64()
        v.data = msg
        return v

    def img_callback(self, data):
        # Получение кадра
        current_frame = self.br.compressed_imgmsg_to_cv2(data)
        success, frame = True, current_frame
        frame = cv2.rotate(frame, cv2.ROTATE_90_COUNTERCLOCKWISE)
        if success:
            channel_count = frame.shape[2]

            # mask = np.zeros(frame.shape, dtype=np.uint8)
            # ignore_mask_color = (255,)*channel_count             
            # cv2.fillPoly(mask, self.roi_corners, ignore_mask_color)
            # masked_image = cv2.bitwise_and(frame, mask)
            masked_image = frame
            

            height, width = frame.shape[0:2]
            pix_per_angle = width / self.fov
            # Нулевые параметры управления
            med_x = width / 2
            controlX = 0.0
            dist_to_gate = 6
            course_to_gate = 0
            
            # Создание HSV маски
            hsv = cv2.cvtColor(masked_image, cv2.COLOR_RGB2HSV)
            # blue = cv2.inRange(hsv, (0, 76, 165), (28, 255, 255))  # -> выделяем синий
            yellow = cv2.inRange(hsv, self.hsv_min, self.hsv_max)  # -> выделяем желтый

            # both = blue + yellow
            both = yellow
            binary = cv2.bitwise_and(both, both, mask=both)
            binary = cv2.morphologyEx(binary, cv2.MORPH_CLOSE, np.ones((5,5),np.uint8))


            # Выделение самых больших контуров удовлетворяющих условию размера
            contours, _ = cv2.findContours(binary, cv2.RETR_EXTERNAL,
                                           cv2.CHAIN_APPROX_NONE)
            contours = list(filter(lambda x: cv2.moments(x)["m00"] > self.min_countur_size, sorted(contours, key=cv2.contourArea, reverse=True)[:2]))
            approxs = list(sorted(contours, key=lambda x: len(cv2.approxPolyDP( 
                        x, self.k * cv2.arcLength(x, True), True)), reverse=True))[:1]
            
            if approxs:
                # Для выделение левого и правого шара
                aligns = []
                # Для выделение ближайшего и дальнего шара
                distances = []
                
                for contour in approxs:
                    contour = cv2.approxPolyDP(contour, self.k * cv2.arcLength(contour, True), True)
                    moments = cv2.moments(contour)
                    # # Получение хз чего
                    # area = cv2.contourArea(contour)
                    # Координаты центра контура
                    cx = int(moments["m10"] / moments["m00"])
                    cy = int(moments["m01"] / moments["m00"])
                    
                    # Вычисление дистанции
                    x, y, w, h = cv2.boundingRect(contour)
                    dist = (w*(self.k1)+self.k2)*0.01


                    # Запись центра контура по горизонту и расчётной дистанции до него
                    aligns.append(cx)
                    distances.append(dist)

                    if True:  # Вывод информации о контуре
                        # Координаты центра
                        # cv2.putText(frame, f'cx: {cx}, cy: {cy}', (cx + 10, cy - 10), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 2, cv2.LINE_AA)
                        # Выделение контура
                        cv2.drawContours(frame, [contour], -1, (0, 255, 0), 1)
                        # Выделение прямоугольника 
                        # cv2.rectangle(frame, (x, y), (x+w, y+h), (0, 255, 0), 2)
                        # Оси через центр контура
                        # cv2.line(frame, (cx, 0), (cx, height), (0, 255, 0), 1)
                        # cv2.line(frame, (0, cy), (width, cy), (0, 255, 0), 1)
                        # Вывести ширину контура
                        cv2.putText(frame, f'W{round(w,2)}', (x, y+h+25), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 0, 255), 2)
                        # Вывести расчётную дистанцию до контура
                        cv2.putText(frame, f'D{round(dist,2)}', (x, y), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 0, 255), 2)
                
                # Расчёт курса и дистанции до ворот
                if aligns:
                    med_x = sum(aligns) / len(aligns)
                controlX = round(-2 * (med_x - width / 2) / width, 1)
                controlX = max(-0.5, min(0.5,controlX))
                course_to_gate = round(controlX * self.fov / 2)
                if distances:
                    dist_to_gate = round(sum(distances) / len(distances), 1)
            
            # Создание отладочного изображения
            miniBin = cv2.resize(binary, (int(binary.shape[1] / 4), int(binary.shape[0] / 4)),
                                 interpolation=cv2.INTER_AREA)
            miniBin = cv2.cvtColor(miniBin, cv2.COLOR_GRAY2RGB)    
            frame[-2 - miniBin.shape[0]:-2, 2:2 + miniBin.shape[1]] = miniBin     
            
            cv2.putText(frame, 'iSee: {};'.format(len(contours)), (width - 120, height - 5),
                        cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 1, cv2.LINE_AA)
            cv2.putText(frame, f'x: {round(dist_to_gate * cos(radians(course_to_gate)), 1)} y: {round(dist_to_gate * sin(radians(course_to_gate)), 1)}', (0, 25),
                        cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 1, cv2.LINE_AA)
            # cv2.putText(frame, f'{course_to_gate = }', (0, 50),
            #             cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 1, cv2.LINE_AA)
            cv2.putText(frame, f'{controlX = }', (0, 75),
                    cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 1, cv2.LINE_AA)

            # Управляющий сигнал
            if abs(controlX) >= 0.01:
                self.z_pub.publish(self.float_to_float64(controlX))
            # Отладочное изображение
            self.pub.publish(self.br.cv2_to_compressed_imgmsg(frame))

        """
        # Получение кадра
        current_frame = self.br.compressed_imgmsg_to_cv2(data)
        success, frame = True, current_frame
        if success:

            height, width = frame.shape[0:2]
            pix_per_angle = width / self.fov
            # Нулевые параметры управления
            controlX = 0.0
            dist_to_gate = 6
            course_to_gate = 0
            
            # Создание HSV маски
            hsv = cv2.cvtColor(frame, cv2.COLOR_RGB2HSV)
            binary = cv2.inRange(hsv, self.hsv_min, self.hsv_max)
            # binary = cv2.inRange(hsv, (20, 76, 139), (37, 175, 255))
            # binary = cv2.inRange(hsv, (25, 55, 161), (45, 255, 255))

            # Выделение самых больших контуров удовлетворяющих условию размера
            contours, _ = cv2.findContours(binary, cv2.RETR_EXTERNAL,
                                           cv2.CHAIN_APPROX_NONE)
            contours = list(filter(lambda x: cv2.moments(x)["m00"] > self.min_countur_size, sorted(contours, key=cv2.contourArea, reverse=True)[:2]))
            
            if len(contours) == 2:
                # Для выделение левого и правого шара
                aligns = []
                # Для выделение ближайшего и дальнего шара
                distances = []
                for contour in contours:
                    # Получение маски контура
                    moments = cv2.moments(contour)
                    # # Получение хз чего
                    # area = cv2.contourArea(contour)
                    # Координаты центра контура
                    cx = int(moments["m10"] / moments["m00"])
                    cy = int(moments["m01"] / moments["m00"])
                    
                    # Вычисление дистанции
                    x, y, w, h = cv2.boundingRect(contour)
                    dist = (w*(self.k1)+self.k2)*0.01

                    # Запись центра контура по горизонту и расчётной дистанции до него
                    aligns.append(cx)
                    distances.append(dist)

                    if True:  # Вывод информации о контуре
                        # Координаты центра
                        # cv2.putText(frame, f'cx: {cx}, cy: {cy}', (cx + 10, cy - 10), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 2, cv2.LINE_AA)
                        # Выделение контура
                        cv2.drawContours(frame, contour, -1, (0, 255, 0), 1)
                        # Выделение прямоугольника 
                        # cv2.rectangle(frame, (x, y), (x+w, y+h), (0, 255, 0), 2)
                        # Оси через центр контура
                        # cv2.line(frame, (cx, 0), (cx, height), (0, 255, 0), 1)
                        # cv2.line(frame, (0, cy), (width, cy), (0, 255, 0), 1)
                        # Вывести ширину контура
                        cv2.putText(frame, f'W{round(w,2)}', (x, y+h+25), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 0, 255), 2)
                        # Вывести расчётную дистанцию до контура
                        cv2.putText(frame, f'D{round(dist,2)}', (x, y), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 0, 255), 2)
                
                # Расчёт курса и дистанции до ворот
                med_x = sum(aligns) / len(aligns)
                controlX = round(-2 * (med_x - width / 2) / width, 1)
                course_to_gate = round(controlX * self.fov / 2)
                dist_to_gate = round(sum(distances) / len(distances), 1)
            
            # Создание отладочного изображения
            miniBin = cv2.resize(binary, (int(binary.shape[1] / 4), int(binary.shape[0] / 4)),
                                 interpolation=cv2.INTER_AREA)
            miniBin = cv2.cvtColor(miniBin, cv2.COLOR_GRAY2RGB)    
            frame[-2 - miniBin.shape[0]:-2, 2:2 + miniBin.shape[1]] = miniBin     
            
            # cv2.putText(frame, 'iSee: {};'.format(len(contours)), (width - 120, height - 5),
            #             cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 1, cv2.LINE_AA)
            cv2.putText(frame, f'{dist_to_gate = }', (0, 25),
                        cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 1, cv2.LINE_AA)
            cv2.putText(frame, f'{course_to_gate = }', (0, 50),
                        cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 1, cv2.LINE_AA)
            cv2.putText(frame, f'{controlX = }', (0, 75),
                    cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 0, 0), 1, cv2.LINE_AA)

            # Управляющий сигнал
            self.z_pub.publish(self.float_to_float64(max(-0.5, min(0.5,controlX))))
            # Отладочное изображение
            self.pub.publish(self.br.cv2_to_compressed_imgmsg(frame))
        """


def main(args=None):
    rclpy.init(args=args)
    RealOpenCV = RealsenseToBinary()
    rclpy.spin(RealOpenCV)
    RealOpenCV.destroy_node()
    rclpy.shutdown()

  
if __name__ == '__main__':
  main()
