#! /usr/bin/env python3

import rclpy
from rclpy.node import Node 
from sensor_msgs.msg import CompressedImage
from std_msgs.msg import Float64, Int16
from geometry_msgs.msg import Point
import cv2 
from cv_bridge import CvBridge
import numpy as np
from math import radians, cos, sin


class RealsenseToBinary(Node):
    def __init__(self):
        super().__init__('RealOpenCV')

        # Фронтальная камера для опознавания ворот
        self.declare_parameter('topic_camera', rclpy.Parameter.Type.STRING)

        # Коэффициенты для расчёта расстояния
        # см. ~/Desktop/python/calc.py
        self.declare_parameter('k1', rclpy.Parameter.Type.DOUBLE) 
        self.declare_parameter('k2', rclpy.Parameter.Type.DOUBLE)
        # Горизонтальный угол обзора камеры
        self.declare_parameter('fov', rclpy.Parameter.Type.INTEGER)
        # 
        self.declare_parameter('gate_position', rclpy.Parameter.Type.STRING)
        self.declare_parameter('blue_gate_position', rclpy.Parameter.Type.STRING)
        self.declare_parameter('heading', rclpy.Parameter.Type.STRING)

        # Параметры для выделения маски
        # yellow = cv2.inRange(hsv, (60, 22, 164), (83, 255, 255)) # -> выделяем желтый
        self.hsv_min = (60, 22, 164)
        self.hsv_max = (83, 255, 255)
        self.min_countur_size = 600
        self.get_logger().info(f"HSV \n{self.hsv_min} \n {self.hsv_max}")

        # Магические числа
        x1 = 0
        x2 = 510
        y1 = 170
        y2 = 280

        self.roi_corners = np.array([[(x1,y1),  (x2,y1), 
                         (x2,y2), (x1,y2),
                       ]], 
                       dtype=np.int32)

        # Параметры для рассчёта расстояние и угла
        self.heading = 0
        self.k1 = self.get_parameter('k1').value
        self.k2 = self.get_parameter('k2').value
        self.fov = self.get_parameter('fov').value

        topic_name = self.get_declare_parameter("topic_camera", lambda x: x, "topic_camera name must be string")
        
        # Топик с отладочным изображением
        self.pub = self.create_publisher(CompressedImage, f"cv/{topic_name}" , 10)
        # Топик с управляющим сигналом
        # self.z_pub = self.create_publisher(Float64, f"{topic_name}/z" , 10)
        self.pub_yellow_pos = self.create_publisher(Point, self.get_declare_parameter("gate_position", lambda x: x, "gate_position topic name must be string"),  10)
        self.pub_blue_pos = self.create_publisher(Point, self.get_declare_parameter("blue_gate_position", lambda x: x, "blue_gate_position topic name must be string"),  10)
        # self.distance_publish = self.create_publisher(Float64, 'distance', 10)
        
        self.br = CvBridge() 
        self.create_subscription(CompressedImage, topic_name, self.img_callback, 10)
        self.create_subscription(Int16, self.get_declare_parameter("heading", lambda x: x, "heading topic name must be string"), self.heading_callback, 10)
        self.get_logger().info(f"CV Start {topic_name}")

    def img_callback(self, data):
        # Получение кадра
        current_frame = self.br.compressed_imgmsg_to_cv2(data)
        success, frame = True, current_frame
        if success:
            channel_count = frame.shape[2]

            # mask = np.zeros(frame.shape, dtype=np.uint8)
            # ignore_mask_color = (255,)*channel_count             
            # cv2.fillPoly(mask, self.roi_corners, ignore_mask_color)
            # masked_image = cv2.bitwise_and(frame, mask)
            masked_image = frame
            height, width = frame.shape[0:2]
            
            # Создание HSV маски
            hsv = cv2.cvtColor(masked_image, cv2.COLOR_RGB2HSV)
            # blue = cv2.inRange(hsv, (0, 76, 165), (28, 255, 255))  # -> выделяем синий
            yellow = cv2.inRange(hsv, (43, 57, 203), (86, 255, 242))  # -> выделяем желтый
            blue = cv2.inRange(hsv, (0, 180, 190), (24, 255, 255)) # -> выделяем синий

            both = blue + yellow
            # both = yellow
            yellow_binary = cv2.bitwise_and(yellow, yellow, mask=yellow)
            blue_binary = cv2.bitwise_and(blue, blue, mask=blue)
            binary = cv2.bitwise_and(both, both, mask=both)

            # Выделение самых больших контуров удовлетворяющих условию размера
            res = self.get_gate(yellow_binary, height, width, frame)
            if res is not None:
                course_to_gate, dist_to_gate = res
                x = round(dist_to_gate * cos(radians(self.heading - course_to_gate)), 1)
                y = round(dist_to_gate * sin(radians(self.heading - course_to_gate)), 1) 
                # Управляющий сигнал
                if 6 >= dist_to_gate >= 1.5:
                    msg = Point()
                    msg.x = x
                    msg.y = y
                    self.pub_yellow_pos.publish(msg)
            
            res = self.get_gate(blue_binary, height, width, frame)
            if res is not None:
                course_to_gate, dist_to_gate = res
                x = round(dist_to_gate * cos(radians(self.heading - course_to_gate)), 1)
                y = round(dist_to_gate * sin(radians(self.heading - course_to_gate)), 1) 
                # Управляющий сигнал
                if 6 >= dist_to_gate >= 1.5:
                    msg = Point()
                    msg.x = x
                    msg.y = y
                    self.pub_blue_pos.publish(msg)

            # Создание отладочного изображения
            miniBin = cv2.resize(binary, (int(binary.shape[1] / 4), int(binary.shape[0] / 4)),
                                 interpolation=cv2.INTER_AREA)
            miniBin = cv2.cvtColor(miniBin, cv2.COLOR_GRAY2RGB)    
            frame[-2 - miniBin.shape[0]:-2, 2:2 + miniBin.shape[1]] = miniBin     
            
            # cv2.putText(frame, 'iSee: {};'.format(len(contours)), (width - 120, height - 5),
            #             cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 1, cv2.LINE_AA)
            # cv2.putText(frame, f'x: {x} y: {y}', (0, 25),
            #             cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 1, cv2.LINE_AA)
            # cv2.putText(frame, f'course: {self.heading - course_to_gate} dist {dist_to_gate}', (0, 50),
            #             cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 1, cv2.LINE_AA)
            # cv2.putText(frame, f'{controlX = }', (0, 75),
            #         cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 1, cv2.LINE_AA)

            # Управляющий сигнал
            # if abs(controlX) >= 0.01:
            #     self.z_pub.publish(self.float_to_float64(controlX))
            # Отладочное изображение
            self.pub.publish(self.br.cv2_to_compressed_imgmsg(frame))

    def get_gate(self, binary, height, width, frame):
        pix_per_angle = width / self.fov
        # Нулевые параметры управления
        med_x = width / 2
        controlX = 0.0
        dist_to_gate = 0
        course_to_gate = 0


        contours, _ = cv2.findContours(binary, cv2.RETR_EXTERNAL,
                                           cv2.CHAIN_APPROX_NONE)
        contours = list(filter(lambda x: cv2.moments(x)["m00"] > self.min_countur_size, sorted(contours, key=cv2.contourArea, reverse=True)[:3]))
        approxs = list(sorted(contours, key=lambda x: len(cv2.approxPolyDP( 
                       x, 0.01 * cv2.arcLength(x, True), True)), reverse=True))[:2]
        
        if len(approxs) == 2:
            # Для выделение левого и правого шара
            aligns = []
            # Для выделение ближайшего и дальнего шара
            distances = []
            for contour in approxs:
                # Точки контура не приближаются к границе ближе чем на 10 пикселей
                # self.get_logger().info(f"{contour}")
                # 1/0
                if any((not 10 < xy[0][0] < width - 10 or not 10 < xy[0][1] < height - 10) for xy in contour):
                    # self.get_logger().info("collide border")
                    continue
                moments = cv2.moments(contour)
                # # Получение хз чего
                # area = cv2.contourArea(contour)
                # Координаты центра контура
                cx = int(moments["m10"] / moments["m00"])
                cy = int(moments["m01"] / moments["m00"])
                
                # Вычисление дистанции
                x, y, w, h = cv2.boundingRect(contour)
                dist = (w*(self.k1)+self.k2)*0.01


                # Запись центра контура по горизонту и расчётной дистанции до него
                aligns.append(cx)
                distances.append(dist)

                if True:  # Вывод информации о контуре
                    # Координаты центра
                    # cv2.putText(frame, f'cx: {cx}, cy: {cy}', (cx + 10, cy - 10), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (255, 255, 255), 2, cv2.LINE_AA)
                    # Выделение контура
                    cv2.drawContours(frame, contour, -1, (0, 255, 0), 1)
                    # Выделение прямоугольника 
                    # cv2.rectangle(frame, (x, y), (x+w, y+h), (0, 255, 0), 2)
                    # Оси через центр контура
                    # cv2.line(frame, (cx, 0), (cx, height), (0, 255, 0), 1)
                    # cv2.line(frame, (0, cy), (width, cy), (0, 255, 0), 1)
                    # Вывести ширину контура
                    cv2.putText(frame, f'W{round(w,2)}', (x, y+h+25), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 0, 255), 2)
                    # Вывести расчётную дистанцию до контура
                    cv2.putText(frame, f'D{round(dist,2)}', (x, y), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 0, 255), 2)
            
            # Расчёт курса и дистанции до ворот
            if len(aligns) == 2:
                med_x = sum(aligns) / len(aligns)
                controlX = round(-2 * (med_x - width / 2) / width, 1)
                controlX = max(-0.5, min(0.5,controlX))
                course_to_gate = round(controlX * self.fov / 2)
                dist_to_gate = round(sum(distances) / len(distances), 1)
                return course_to_gate, dist_to_gate
            else:
                return None
            

    def get_declare_parameter(self, name, checker, error_text):
        value = self.get_parameter(name).value
        if not checker(value):
            self.get_logger().error(error_text)
            raise Exception(error_text)
        return value

    def float_to_float64(self, msg):
        v = Float64()
        v.data = msg
        return v

    def heading_callback(self, msg):
        self.heading = msg.data


def main(args=None):
    rclpy.init(args=args)
    RealOpenCV = RealsenseToBinary()
    rclpy.spin(RealOpenCV)
    RealOpenCV.destroy_node()
    rclpy.shutdown()

  
if __name__ == '__main__':
  main()
